;************************************************************************
;*
;*	Filename : sfmpk.x65
;*	Version  : 1.0
;*	(c) 1993 Nintendo of America Inc.
;*	Translation : Shoya Tanaka
;*
;************************************************************************
;
;
;	Super NES multi process karnel
;
		LIB	DEFINE\65816.DEF
		LIB	DEFINE\SFC.DEF
;
PAGE0		GROUP
		EXTERN	nmi_control
;
process_link	EQU	1
process_address	EQU	3
event_link	EQU	0
;
PAGE0		GROUP
ready_process	DS	2
;
DATA		GROUP
field_event	DS	2
;
CODE		GROUP
		PUBLIC	Init_process
Init_process	mw
		STZ	.ready_process
		STZ	!field_event
		RTL
;
		PUBLIC	Fork_process
Fork_process	xw				;X = process structure
		PLY				;<<
		STY	.process_address+0,X	;w
		mB
		PLA				;<
		STA	.process_address+2,X	;b
		mW
		LDY	.ready_process		;last ready process
		STX	.ready_process		;new process
		BEQ	fork_first
		LDA	!process_link,Y		;1st ready process
		STX	.process_link,Y		;new process
		STA	.process_link,X
		RTL
fork_first	TXY
		STX	.process_link,Y		;new process
		RTL				;<<<
;
		PUBLIC	Post_field
Post_field	mwxw
		LDX	#field_event
;
		PUBLIC	Post_event
Post_event	mwxw				;X = event structure
		LDY	.event_link,X		;last wait process
		BEQ	post_return
		STZ	.event_link,X
		LDX	.ready_process		;last ready process
		STY	.ready_process
		BEQ	post_return
		LDA	!process_link,Y		;1st wait process
		LDY	.process_link,X		;1st ready process
		STA	.process_link,X
		LDX	.ready_process		;last wait process
		STY	.process_link,X
post_return	RTL
;
		PUBLIC	Wait_field
Wait_field	mwxw
		LDX	#field_event
;
		PUBLIC	Wait_event
Wait_event	mwxw				;X = event structure
		PHA				;>>
		TSC				;new wait process
 		LDY	.event_link,X		;last wait process
		STA	.event_link,X
		BEQ	wait_skip
		LDA	!process_link,Y		;1st wait process
		TSX
		STX	.process_link,Y
wait_skip	STA	.process_link,S
;
		PUBLIC	Join_process
Join_process	mwxw
dispatch	LDX	.ready_process		;last READY process
		BNE	dispatch_entry
		PHP				;>
		mBxB
		LDA	#enable_NMI		;b
		TSB	.nmi_control		;b
		SEI
		LDX	.nmi_control		;b
		BIT	!NMI_status		;b
		STX	!NMI_control		;b
		PLP				;<
		mwxw
		PHP				;>
		mB
dispatch_idle	LDX	.ready_process		;last READY process
		BEQ	dispatch_idle
		TRB	.nmi_control		;b
		SEI
		LDA	.nmi_control		;b
		STA	!NMI_control		;b
		PLP				;<
		mwxw
dispatch_entry	LDA	.process_link,X		;1st READY process
		TCS
		PLY				;<< 2nd READY process
		STY	.process_link,X
		CMP	.ready_process		;last == 1st
		BNE	dispatch_skip
		STZ	.ready_process
dispatch_skip	RTL				;<<<
;
		END

	Super NES multi process kernel (V1.02)

0. initialize multiprocess system

Initiate_process is called, then system is initialized. This means, executing
program is considered the root process, and the root process is running.

[example]
CODE		GROUP
		 |
		REP	#%00010000
		JSL	Initiate_process
		 |


1. Create process

Stack area for process and following process control structure (6 bytes RAM)
are defined, and the process control structure address is designated by X
register, and Fork_process is called, then child process is excuted. The 
child process is progarm which begin from program after "JSR Fork_process".
Process which call Fork_process continue, so it means to gain one process.
The process which was executed become preparatory condition once, and it begin
to run really after other high priority process became stoped or waitting.

[example]
DATA		GROUP
		DS	26		;define stack area for process
sample_process	DS	6		;define process control structure

CODE		GROUP
		 |			;owner process
		JSL	Sample_process
		 |

Sample_process	REP	#%00110000	;child process
		LDX	#sample_process
		JSL	Fork_process
		 |
		child process program
		 |


2. Wait event

Event control structure (2 bytes RAM) is defined, and the address is set to
X-register, and Wait_event is called, then "Wait event" is available.
After execution, the process which call this process goes to a
waitting condition.

[example]
DATA		GROUP
sample_event	DW	1		;define event structure

CODE		GROUP
		 |
		REP	#%00110000
		LDX	#sample_event
		JSL	Wait_event
		 |



3. Event occur

For interuption disposal e.t.c., event control structure address is designated
by X-register, and Post_event is called, then event is able to be occured.
All processes that are waitting this event become ready condition to excute.

[example]
CODE		GROUP
		 |
		REP	#%00110000
		LDX	#sample_event
		JSL	Post_event
		 |


4. Stop process

It is possibe to stop process by Stop_process. Process became stopping condition.
If all processes are stopped, it is impossible to create new process. Pay
attention abvoe situation.

[example]
CODE		GROUP
		 |
		REP	#%00110000
		JML	Join_process


5. Wait field (frame)

If you want process to be excuted with synchrons to the screen scanning rate
(1/60 sec), you can synchroniz by using the supervisor Wait_field.
The instruction after this supervisor is not executed until next vertical
blanking (NMI). Pay attention that the Wait field may not necessarily be excuted
during the vertical blanking period.
In order to execute this supervisor, supervisor Post_field must be executed 
once during NMI disposal.

[example]
CODE		GROUP
		 |
		REP	#%00110000
		JSL	Wait_field
		 |

BANK0		GROUP
NMI_native	PHB				;>
		PHK				;>
		PLB				;<
		REP	#%00110000
		PHA				;>>
		PHX				;>>
		PHY				;>>
		LDA	#0
		TCD
		 |
		JSL	Post_field
		 |
		PLY				;<<
		PLX				;<<
		PLA				;<<
		PLB				;<
		RTI				;<<<<

COMMON		GROUP
		ORG	$FFEA
		DW	NMI_native


6. Caution

When several processes are executed at the same time, Each process execution
needs sutiable time given by Wait_event. Otherwise, other processes will not be 
executed. In short, other processes are executed during Wait_event command.

In several peocesses, their turn to run is equal to become ready conditon.
Processes that became ready condition at the same time will run by turn
to become wait condition.

One process can not stop another process. There is not function to stop as a
whole in the relation between owner process and child peocess.

To control NMI, it is necessary to define register (NMI_control etc) in outside
and to set. NMI is enable when there is no process for execution, so NMI should
be set to disable usually. Futher, NMI must be disable in every part including
IRQ.

When new event structure is defined, it is nessesary to clear by "0" in the
content.

7. Process situation transition


	Fork_process		Post_event


		____________________
	>	|	ready      |------------
       |        --------------------           |
       |                                       v
____________________                    __________________
|	stop       |			|	wait     |
--------------------                    ------------------
       ^                                       |
       |                        dispatch       |
       |        ____________________           |
       ---------|	run        |      <-----
                --------------------

	Join_process		Wait_event


8. Process control structure

Ready or waitting stack top is process control structure. 
Process control structure remembers process which are linked ,and remembers 
their return address from supervisor.


	RAEDY or WAIT process
	--------------------------
	stack area (unuse)
		process structure
	--------------------------
	process link (2byte)
	--------------------------
	return address (3byte)
	--------------------------
	stack area (use)
	--------------------------

Register Ready_process designate ready process, that is process's circulatory
list end that is possible to execute.

		1ST READY_PROCESS	2ND READY_PROCESS

	------>	stack area	------>	stack
	|	process link----|	process link-----
	|	return address		return address	|
	|	stack area		stack area	|
	|						|
        -------------------------------------------------




Event control structure designate waitting process, that is process's 
circulatory list end that process is waittng the event.

		event structure	;shoya omit below, see printed paper.


9. Requrement and condition when process supervisor is in use.

Show requrement when each supervisor is called, and returned condition.



About above figure, - means optional, ? means unfixed, up < means preservation,
P means process structure, E means event structure. * means to return one nest
upper subroutine.

