;************************************************************************
;*
;*	Filename : sprite.x65
;*	Version  : 1.0
;*	(c) 1993 Nintendo of America Inc.
;*	Translation : Shoya Tanaka
;*
;************************************************************************
;
;
;	Sprite control process
;
		LIB	DEFINE\65816.DEF
		LIB	DEFINE\SFC.DEF
;
		PUBLIC	Sprite_process,sprite_buffer,sprite_pointer
CODE		GROUP
		EXTERN	Set_sprite
		EXTERN	Fork_process,Wait_field,Reserve_NMI
;
PAGE0		GROUP
sprite_pointer	DS	2
;
DATA		GROUP
		DS	26		;process control block
sprite_process	DS	6
sprite_execute	DS	5		;reservation execution control block
sprite_buffer	DS	512		;main sprite buffer
		DS	32		;sub sprite buffer
		DS	512		;lined sub buffer
;
SPRITE_BEGIN	EQU	0
SPRITE_END	EQU	512
;
CODE		GROUP
;	DMA transfer all sprite	
		mbxw
Execute_sprite	LDX	#sprite_buffer			;w
		STX	!DMA_1+DMA_address+0		;w
		LDX	#512+32				;w
		STX	!DMA_1+DMA_count		;w
		STZ	!Sprite_address+0		;b
		STZ	!Sprite_address+1		;b
		LDX	#sprite_execute			;w
		JSL	~Reserve_NMI
		LDA	#dma_1				;b
		STA	!DMA_burst			;b
		RTL
;
;	Sprite process
		mwxw
Sprite_process	LDX	#sprite_process			;w
		JSL	~Fork_process
		mB
		LDA	#dma_ST+dma_INC+dma_1W		;b
		STA	!DMA_1+DMA_control		;b
		LDA	#Sprite_write&$FF		;b
		STA	!DMA_1+DMA_register		;b
		STZ	!DMA_1+DMA_address+2		;b
;
;	Define dynamic sprite
		mW
sprite_loop	LDX	#sprite_buffer+SPRITE_BEGIN	;w
		STX	.sprite_pointer			;w
		JSL	~Set_sprite
;
;	Delete surplus dynamic sprite
		mB
sprite_clear	LDX	.sprite_pointer			;w
		BRA	_entry
_loop		LDA	#$80				;b
		STA	.0,X				;b
		LDA	#%00000001			;b
		STA	!512+32,X			;b
		INX
		INX
		INX
		INX
_entry		CPX	#sprite_buffer+SPRITE_END	;w
		BCC	_loop
;
;	Sort sub data of all sprite
sprite_arrange	LDX	#sprite_buffer+512+32+512	;w
		LDY	#32				;w
_loop		mW	fC
		TXA
		SBC	#15				;w
		TAX
		mB
		LDA	.12,X				;b
		ASL	A
		ASL	A
		ORA	.08,X				;b
		ASL	A
		ASL	A
		ORA	.04,X				;b
		ASL	A
		ASL	A
		ORA	.00,X				;b
		DEY
		STA	!sprite_buffer+512,Y		;b
		BNE	_loop
;
;	Reserve sprite transfer disposal
		JSL	~Execute_sprite
		mW
		JSL	~Wait_field
		JMP	sprite_loop
;
		END
;
			Sprite control process

The maximum number of OBJ's that can be displayed on the screen is 128 at the
Super NES. This process control OBJ like following.

First, sprite is grouped into static sprite and dynamic sprite. Static sprite
means sprite that parameter does not change with time or a few parameter
change. Static slprite is not object of control virtually. Dynamic sprite
means sprite that almost all parameter change or priority change with time,
and Dynamic sprite is controlled by the program.

In sprite area (128 sprites), Static sprite is assigned to the top and end of 
the sprite area, and Dynamic is assigned to the middle of the sprite area.



Sprite buffer layout

Sprite that is displayed before Dynamic sprite (higher priority sprite)
is assigned in the top of the area, sprite that is displayed behind Dynamic
sprite (lower priority sprite) is assigned in the end of this area. In the
middle of this area, Dynamic sprite that is displayed is assigned to the area
first, and Dynamic sprite thta is not displayed is assigned to the rest area.

Sprite buffer is expressed in main buffer 4 bytes and sub buffer 2 bits.
It is difficult location to control each sprite, so sub buffer is opened 
in 4 bytes and is assigned to other memory area.

Dynamic sprite is set by Set_sprite that is located outside.

After sprite setting is done, the data in sub buffer is packed to 32 bytes,
and is copied behind main buffer. Main buffer and sub buffer that are made in 
this way are transfered to PPU by DMA channel 1 during vertical blanking period.



